"""Check that any raw df conforms to the required format."""

from typing import Any, Optional, Union

import pandas as pd


def check_for_duplicates(
    df: pd.DataFrame,
    subset_duplicates_columns: Union[list[str], str],
) -> tuple[pd.Series, list[str]]:
    """Check for duplicates in the dataframe.

    Args:
        df (pd.DataFrame): Dataframe to check.
        subset_duplicates_columns (Union[list[str], str]): list of columns to subset on when checking for duplicates.

    Returns:
        tuple(pd.Series, list[str]): duplicates and list of errors.
    """

    source_failures = []

    duplicates_idx = df.duplicated(subset=subset_duplicates_columns, keep=False)

    if duplicates_idx.any():
        source_failures.append(f"Duplicates found on {subset_duplicates_columns}")
        duplicates = df[duplicates_idx]
    else:
        duplicates = None

    return duplicates, source_failures


def get_column_dtype_failures(
    df: pd.DataFrame,
    expected_val_dtypes: list[str],
    col: str,
) -> Union[str, None]:
    """Check that the column is of the correct dtype.

    Args:
        df (pd.DataFrame): Dataframe to check.
        expected_val_dtypes (list[str]): Expected dtype of value column.
        col (str): Column to check.

    Returns:
        str: errors
    """

    if "timestamp" in col:
        # Check that column has a valid datetime format
        if df[col].dtype != "datetime64[ns]":
            return f"{col}: invalid datetime format"
    elif "value" in col:
        # Check that column has a valid numeric format
        if df[col].dtype not in expected_val_dtypes:
            return f"{col}: dtype {df[col].dtype}, expected {expected_val_dtypes}"

    return None


def get_na_prop_failures(
    df: pd.DataFrame,
    allowed_nan_value_prop: float,
    col: str,
) -> Union[str, None]:
    """Check if column has too many missing values.

    Args:
        df (pd.DataFrame): Dataframe to check.
        allowed_nan_value_prop (float): Allowed proportion of missing values.
        col (str): Column to check.

    Returns:
        str: Error message if too many missing values.
    """

    na_prop = df[col].isna().sum() / df.shape[0]

    if na_prop > 0:
        if col != "value":
            return f"{col}: {na_prop} NaN"
        else:
            if na_prop > allowed_nan_value_prop:
                return f"{col}: {na_prop} NaN (allowed {allowed_nan_value_prop})"

    return None


def check_required_columns(
    df: pd.DataFrame,
    required_columns: list[str],
    allowed_nan_value_prop: float,
    expected_val_dtypes: list[str],
) -> list[str]:
    """Check that the required columns are present and that the value column.

    Args:
        df (pd.DataFrame): Dataframe to check.
        required_columns (list[str]): list of required columns.
        allowed_nan_value_prop (float): Allowed proportion of missing values.
        expected_val_dtypes (list[str]): Expected dtype of value column.

    Returns:
        list[str]: list of errors.
    """

    source_failures = []

    for col in required_columns:
        if col not in df.columns:
            source_failures.append(f"{col}: not in columns")
            continue

        # Check that columns are correctly formatted
        dtype_failures = get_column_dtype_failures(
            df=df,
            expected_val_dtypes=expected_val_dtypes,
            col=col,
        )

        # Check for NaN in cols
        na_prop_failures = get_na_prop_failures(
            df=df,
            allowed_nan_value_prop=allowed_nan_value_prop,
            col=col,
        )

        for f_type in dtype_failures, na_prop_failures:
            if f_type:
                source_failures.append(f_type)

    return source_failures


def check_raw_df(  # pylint: disable=too-many-branches
    df: pd.DataFrame,
    required_columns: Optional[list[str]] = None,
    allowed_nan_value_prop: float = 0.0,
    expected_val_dtypes: Optional[list[str]] = None,
    subset_duplicates_columns: Union[list, str] = "all",
    raise_error: bool = True,
) -> tuple[list[str], Any]:
    """Check that the raw df conforms to the required format and doesn't
    contain duplicates or missing values.

    Args:
        df (pd.DataFrame): Dataframe to check.
        required_columns (list[str]): list of required columns. Defaults to ["dw_ek_borger", "timestamp", "value"].
        allowed_nan_value_prop (float): Allowed proportion of missing values. Defaults to 0.0.
        expected_val_dtypes (list[str]): Expected dtype of value column. Defaults to "float64".
        subset_duplicates_columns (Union[list, str]): Which columns to check for duplicates across. Defaults to "all".
        raise_error (bool): Whether to raise an error if the df fails the checks. Defaults to True.

    Returns:
        list[str]: list of errors.

    Raises:
        ValueError: If the df fails the checks and raise_error is True.
    """
    source_failures = []

    if required_columns is None:
        if "value" in df.columns:
            required_columns = ["dw_ek_borger", "timestamp", "value"]
        elif "date_of_birth" in df.columns:
            required_columns = ["dw_ek_borger", "date_of_birth"]
        elif "sex_female" in df.columns:
            required_columns = ["dw_ek_borger", "sex_female"]
        else:
            required_columns = ["dw_ek_borger", "timestamp"]

    if subset_duplicates_columns == "all":
        subset_duplicates_columns = list(df.columns)

    # Raise error if string other than 'All' is passed to subset_duplicates_columns
    if (
        isinstance(subset_duplicates_columns, str)
        and subset_duplicates_columns != "all"
    ):
        raise ValueError(
            "subset_duplicates_columns must be a list of column names or 'All'",
        )

    if expected_val_dtypes is None:
        expected_val_dtypes = ["float64", "int64"]

    # Check that the dataframe has a meaningful length
    if df.shape[0] == 0:
        source_failures.append("No rows returned")

    # Check that the dataframe has the required columns
    source_failures += check_required_columns(
        df=df,
        required_columns=required_columns,
        allowed_nan_value_prop=allowed_nan_value_prop,
        expected_val_dtypes=expected_val_dtypes,
    )

    # Check for duplicates
    duplicates, duplicate_sources = check_for_duplicates(
        df=df,
        subset_duplicates_columns=subset_duplicates_columns,
    )

    source_failures += duplicate_sources

    # Raise error if any failures
    if raise_error and len(source_failures) > 0:
        raise ValueError(source_failures)

    return source_failures, duplicates



"""Loaders for structured SFI-data."""

# pylint: disable = missing-function-docstring

from typing import Optional

import pandas as pd

from psycop_feature_generation.loaders.raw.sql_load import sql_load
from psycop_feature_generation.utils import data_loaders


def sfi_loader(
    aktivitetstypenavn: Optional[str] = None,
    elementledetekst: Optional[str] = None,
    n_rows: Optional[int] = None,
    value_col: str = "numelementvaerdi",
) -> pd.DataFrame:
    """Load structured_sfi data. By default returns entire structured_sfi data
    view with numelementværdi as the value column.

    Args:
        aktivitetstypenavn (str): Type of structured_sfi, e.g. 'broeset_violence_checklist', 'selvmordsvurdering'. Defaults to None. # noqa: DAR102
        elementledetekst (str): elementledetekst which specifies which sub-element of the SFI, e.g. 'Sum', "Selvmordstanker". Defaults to None.
        n_rows: Number of rows to return. Defaults to None which returns entire structured_sfi data view.
        value_col: Column to return as value col. Defaults to 'numelementvaerdi'.

    Returns:
        pd.DataFrame
    """
    view = "[FOR_SFI_uden_fritekst_resultater_psyk_somatik_inkl_2021_feb2022]"
    sql = f"SELECT dw_ek_borger, datotid_resultat_udfoert, {value_col} FROM [fct].{view} WHERE datotid_resultat_udfoert IS NOT NULL"

    if aktivitetstypenavn:
        sql += f" AND aktivitetstypenavn = '{aktivitetstypenavn}'"
    if elementledetekst:
        sql += f" AND elementledetekst = '{elementledetekst}'"

    df = sql_load(sql, database="USR_PS_FORSK", chunksize=None, n_rows=n_rows)

    # Drop duplicate rows
    df = df.drop_duplicates(keep="first")

    # Drop rows with duplicate dw_ek_borger and datotid_resultat_udfoert
    # Data contained rows with scores reported at the same time for the same patient but with different values
    df = df.drop_duplicates(
        subset=["datotid_resultat_udfoert", "dw_ek_borger"],
        keep="first",
    )

    df.rename(
        columns={
            "datotid_resultat_udfoert": "timestamp",
            value_col: "value",
        },
        inplace=True,
    )

    return df.reset_index(drop=True)


@data_loaders.register("broeset_violence_checklist")
def broeset_violence_checklist(n_rows: Optional[int] = None) -> pd.DataFrame:
    return sfi_loader(
        aktivitetstypenavn="Brøset Violence Checkliste (BVC)",
        elementledetekst="Sum",
        n_rows=n_rows,
        value_col="numelementvaerdi",
    )


@data_loaders.register("broeset_violence_checklist_physical_threats")
def broeset_violence_checklist_physical_threats(
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    df = sfi_loader(
        aktivitetstypenavn="Brøset Violence Checkliste (BVC)",
        elementledetekst="Fysiske trusler",
        n_rows=n_rows,
        value_col="elementkode",
    )

    df["value"] = df["value"].replace(
        to_replace=[
            "010BroesetNulPoint",
            "020BroesetEtPoint",
        ],
        value=[0, 1],
        regex=False,
    )

    return df


@data_loaders.register("selvmordsrisiko")
def selvmordsrisiko(n_rows: Optional[int] = None) -> pd.DataFrame:
    df = sfi_loader(
        aktivitetstypenavn="Screening for selvmordsrisiko",
        elementledetekst="ScrSelvmordlRisikoniveauKonkl",
        n_rows=n_rows,
        value_col="elementkode",
    )

    df["value"] = df["value"].replace(
        to_replace=[
            "010ScrSelvmordKonklRisikoniveau1",
            "020ScrSelvmordKonklRisikoniveau2",
            "030ScrSelvmordKonklRisikoniveau3",
        ],
        value=[1, 2, 3],
        regex=False,
    )

    return df


@data_loaders.register("hamilton_d17")
def hamilton_d17(n_rows: Optional[int] = None) -> pd.DataFrame:
    return sfi_loader(
        aktivitetstypenavn="Vurdering af depressionssværhedsgrad med HAM-D17",
        elementledetekst="Samlet score HAM-D17",
        n_rows=n_rows,
        value_col="numelementvaerdi",
    )


@data_loaders.register("mas_m")
def mas_m(n_rows: Optional[int] = None) -> pd.DataFrame:
    return sfi_loader(
        aktivitetstypenavn="MAS-M maniscoringsskema (Modificeret Bech-Rafaelsen Maniskala)",
        elementledetekst="MAS-M score",
        n_rows=n_rows,
        value_col="numelementvaerdi",
    )


@data_loaders.register("height_in_cm")
def height_in_cm(n_rows: Optional[int] = None) -> pd.DataFrame:
    return sfi_loader(
        aktivitetstypenavn="Måling af patienthøjde (cm)",
        elementledetekst="Højde i cm",
        n_rows=n_rows,
        value_col="numelementvaerdi",
    )


@data_loaders.register("weight_in_kg")
def weight_in_kg(n_rows: Optional[int] = None) -> pd.DataFrame:
    return sfi_loader(
        aktivitetstypenavn="Måling af patientvægt (kg)",
        elementledetekst="Vægt i kg",
        n_rows=n_rows,
        value_col="numelementvaerdi",
    )


@data_loaders.register("bmi")
def bmi(n_rows: Optional[int] = None) -> pd.DataFrame:
    return sfi_loader(
        aktivitetstypenavn="Bestemmelse af Body Mass Index (BMI)",
        elementledetekst="BMI",
        n_rows=n_rows,
        value_col="numelementvaerdi",
    )



"""Some of our data series contain both numerical values and strings like
'>=12'.

This function takes a df and coerces all strings into numerical.
"""

from typing import Optional

import pandas as pd


def multiply_inequalities_in_df(
    df: pd.DataFrame,
    round_to_decimals: int = 6,
    ineq2mult: Optional[dict[str, float]] = None,
    col_to_multiply: str = "value",
) -> pd.DataFrame:
    """Multiply inequalities in a dataframe by a factor.

    Args:
        df (pd.Dataframe): The dataframe to be modified.
        round_to_decimals (int): How many decimals to round the value to.
        ineq2mult (dict[str, float]): A dictionary with the inequalities as keys and the factors as values.
            Current values are arbitrary, but ensure that inequalities are somewhat separated from the continuous part of the distribution.
        col_to_multiply (str): The column to multiply.

    Returns:
        pd.Dataframe: The modified dataframe.
    """
    if (
        ineq2mult is None
    ):  # Avoid sharing a mutable keyword argument between function invocations
        ineq2mult = {
            "<": 0.67,
            "<=": 0.8,
            ">": 1.5,
            ">=": 1.2,
        }

    # Sort inequalities by length, so that we don't replace "<" in "<=".
    in_eqs = sorted(ineq2mult.keys(), key=len, reverse=True)

    for in_eq in in_eqs:
        col_can_contain_ineq = df[col_to_multiply].dtype == "object"

        if col_can_contain_ineq:
            starts_with_ineq_idxs = (
                df[col_to_multiply].str.startswith(in_eq).fillna(False)
            )

            df.loc[starts_with_ineq_idxs, col_to_multiply] = (
                df.loc[starts_with_ineq_idxs, col_to_multiply]
                .str.replace(",", ".")
                .str.extract(r"(\d+\.\d+|\d+)", expand=False)
                .astype(float)
                .mul(ineq2mult[in_eq])
                .round(round_to_decimals)
            )

    # Convert col_to_multiply dtype to float
    df[col_to_multiply] = df[col_to_multiply].astype(float)

    return df



"""Example loader for admissions."""
import psycop_feature_generation.loaders.raw as r

if __name__ == "__main__":
    df = r.load_admissions.admissions(n_rows=1000)
    psych = r.load_admissions.admissions_to_psychiatry(n_rows=1000)
    somatic = r.load_admissions.admissions_to_somatic(n_rows=1000)



"""Functions for validating raw data – in the sense of data returned from a
loader."""
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
from deepchecks.tabular import Dataset
from deepchecks.tabular.suites import data_integrity
from wasabi import Printer

from psycop_feature_generation.data_checks.flattened.data_integrity import (
    get_failed_check_names,
)
from psycop_feature_generation.data_checks.flattened.feature_describer import (
    create_unicode_hist,
)
from psycop_feature_generation.data_checks.utils import save_df_to_pretty_html_table


def median_absolute_deviation(series: pd.Series) -> float:
    """Calculates a series' median absolute deviation from its own median.

    Args:
        series (pd.Series): Series to calculate the median absolute deviation of.

    Returns:
        np.array: Median absolute deviation of the series.
    """
    med = np.median(series)
    return np.median(np.abs(series - med))


def generate_column_description(series: pd.Series) -> dict:
    """Generates a dictionary with column description.

    Args:
        series (pd.Series): Series to describe.

    Returns:
        dict: dictionary with column description.
    """

    d = {
        "col_name": series.name,
        "dtype": series.dtype,
        "n_unique": series.nunique(),
        "n_missing": series.isna().sum(),
        "min": series.min(),
        "max": series.max(),
        "mean": series.mean(),
        "std": series.std(),
        "median": series.median(),
        "median_absolute_deviation": median_absolute_deviation(series),
    }
    d["histogram"] = create_unicode_hist(series)
    for percentile in [0.01, 0.25, 0.5, 0.75, 0.99]:
        d[f"{percentile}th_percentile"] = round(series.quantile(percentile), 1)

    return d


def get_na_prob(series: pd.Series) -> pd.Series:
    """Calculates the propotion of rows that are NaT.

    Args:
        series (pd.Series): Series of timestamps

    Returns:
        pd.Series: Series with proportion of NaT
    """
    return series.isna().sum() / len(series)


def highlight_large_deviation(
    series: pd.Series,
    threshold_ratio: float,
    baseline_column: str,
    variation_column: str,
) -> list[str]:
    """Highlights rows where the 99th/1st percentile is x times the variation
    column larger/smaller than the baseline column (probably mean or median).

    Args:
        series (pd.Series): Series to describe.
        threshold_ratio (float): Threshold for deviation. 3-4 might be a good value.
        baseline_column (str): Name of the column to use as baseline. Commonly 'mean' or 'median'.
        variation_column (str): Name of the column containing the variation.
        Commonly 'std' or 'median_absolute_deviation'

    Returns:
        list[str]: list of styles for each row.
    """
    above_threshold = pd.Series(data=False, index=series.index)
    lower_bound = series[baseline_column] - series[variation_column] * threshold_ratio
    upper_bound = series[baseline_column] + series[variation_column] * threshold_ratio

    above_threshold[baseline_column] = (
        series.loc["0.99th_percentile"] > upper_bound
        or series.loc["0.01th_percentile"] < lower_bound
    )
    return [
        "background-color: yellow" if above_threshold.any() else ""
        for v in above_threshold
    ]


def validate_raw_data(
    save_path: Path,
    df: pd.DataFrame,
    feature_set_name: str,
    deviation_baseline_column: Optional[str] = "median",
    deviation_threshold_ratio: Optional[float] = 4.0,
    deviation_variation_column: Optional[str] = "median_absolute_deviation",
) -> None:
    """Validates raw data from SQL database (or any dataframe, really). Runs
    data integrity checks from deepchecks, and calculates summary statistics.
    Summary statistics are saved as a table with one row for each column. Rows
    are colored yellow if the 99th/1st percentile exceeds.

    `deviation_baseline_column'  +- `deviation_treshold` * `deviation_variation_column`.
    All files are saved to the `RAW_DATA_VALIDATION_PATH` directory in a subdirectory
    named `feature_set_name`.

    Args:
        save_path (Path): Path to save the validation results to.
        df (pd.DataFrame): Dataframe to validate.
        feature_set_name (str): Name of the feature set.
        deviation_baseline_column (Optional[str], optional): Which metric to use as
            the baseline for identifying columns that might be problematic. Defaults to "median".
        deviation_threshold_ratio (Optional[float], optional): Which threshold to
            use as the cutoff for identifying problematic columns. Defaults to 4.0.
        deviation_variation_column (Optional[str], optional): Which metric to use
            as the measure of variation from the baseline when identifying problematic
            columns. Defaults to "median_absolute_deviation".

    Raises:
        ValueError: If 'dw_ek_borger' or 'timestamp' are not columns in the dataframe.
    """

    msg = Printer(timestamp=True)
    failed_checks = {}

    save_path.mkdir(parents=True, exist_ok=True)

    # check if `timestamp` and `dw_ek_borger` columns exist
    timestamp_col_name = "timestamp" if "timestamp" in df.columns else None
    id_col_name = "dw_ek_borger" if "dw_ek_borger" in df.columns else None
    if timestamp_col_name or id_col_name is None:
        raise ValueError(
            "Dataframe must contain `timestamp` and `dw_ek_borger` columns.",
        )

    # Deepchecks
    d_set = Dataset(
        df=df,
        index_name=id_col_name,
        datetime_name=timestamp_col_name,
        cat_features=[],
    )
    integ_suite = data_integrity(timeout=0)

    with msg.loading("Running data integrity checks..."):
        suite_results = integ_suite.run(d_set)
        suite_results.save_as_html(str(save_path / "data_integrity.html"))
        failed_checks["data_integrity"] = get_failed_check_names(suite_results)

    msg.good("Finished data integrity checks.")

    # Data description
    data_columns = [
        col for col in df.columns if col not in [id_col_name, timestamp_col_name]
    ]
    with msg.loading("Generating data description..."):
        data_description = [
            generate_column_description(df[col]) for col in data_columns
        ]

    msg.good("Finished data description.")

    data_description = pd.DataFrame(data_description)
    data_description["prop NaT"] = get_na_prob(df[timestamp_col_name])  # type: ignore
    data_description.to_csv(save_path / "data_description.csv", index=False)  # type: ignore
    # Highlight rows with large deviations from the baseline
    data_description = data_description.style.apply(  # type: ignore
        func=highlight_large_deviation,
        threshold=deviation_threshold_ratio,
        baseline_column=deviation_baseline_column,
        variation_column=deviation_variation_column,
        axis=1,
    )

    save_df_to_pretty_html_table(
        data_description,
        save_path / "data_description.html",
        title=f"Data description - {feature_set_name}",
        subtitle=f"Yellow rows indicate {deviation_threshold_ratio}x deviations from the {deviation_baseline_column}\n(99th/1st percentile within {deviation_baseline_column} +- {deviation_variation_column} * threshold={deviation_threshold_ratio}) from the baseline.)",
    )

    msg.info(f"All files saved to {save_path}")
    if failed_checks:
        print(
            f"The following checks failed - look through the generated reports!\n{failed_checks}",
        )



"""Utilities for data checks."""

from pathlib import Path
from typing import Optional

import pandas as pd

# Templates for saving dataframes as pretty html tables
HTML_TEMPLATE1 = """
<html>
<head>
<style>
  h2 {
    text-align: center;
    font-family: Helvetica, Arial, sans-serif;
  }
  table { 
    margin-left: auto;
    margin-right: auto;
  }
  table, th, td {
    border: 1px solid black;
    border-collapse: collapse;
  }
  th, td {
    padding: 5px;
    text-align: center;
    font-family: Helvetica, Arial, sans-serif;
    font-size: 90%;
  }
  table tbody tr:hover {
    background-color: #dddddd;
  }
  .wide {
    width: 90%; 
  }
</style>
</head>
<body>
"""

HTML_TEMPLATE2 = """
</body>
</html>
"""


def save_df_to_pretty_html_table(
    df: pd.DataFrame,
    path: Path,
    title: Optional[str] = None,
    subtitle: Optional[str] = None,
) -> None:
    """Write dataframe to a HTML file with nice formatting. Stolen from
    stackoverflow: https://stackoverflow.com/a/52722850.

    Args:
        df (pd.DataFrame): Dataframe to write.
        path (Path): Path to save the HTML file to.
        title (Optional[str], optional): Title for the table. Defaults to None.
        subtitle (Optional[str], optional): Subtitle for the table. Defaults to None.
    """

    html = ""

    if title:
        html += f"<h2> {title} </h2>\n"
    if subtitle:
        html += f"<h3> {subtitle} </h3>\n"
    html += df.to_html(classes="wide", escape=False)

    with open(path, "w", encoding="utf-8") as f:
        f.write(HTML_TEMPLATE1 + html + HTML_TEMPLATE2)



"""Loaders for coercion data."""

# pylint: disable = non-ascii-name,missing-function-docstring

from typing import Optional

import pandas as pd

from psycop_feature_generation.loaders.raw.sql_load import sql_load
from psycop_feature_generation.utils import data_loaders


@data_loaders.register("coercion_duration")
def coercion_duration(
    coercion_type: Optional[str] = None,
    reason_for_coercion: Optional[str] = None,
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Load coercion data. By default returns entire coercion data view with
    duration in hours as the value column.

    Args:
        coercion_type (str): Type of coercion, e.g. 'tvangsindlæggelse', 'bæltefiksering'. Defaults to None. # noqa: DAR102
        reason_for_coercion (str): Reason for coercion, e.g. 'farlighed'. Defaults to None.
        n_rows: Number of rows to return. Defaults to None which returns entire coercion data view.

    Returns:
        pd.DataFrame
    """
    coercion_discard = """('Døraflåsning', 'Personlig afskærmning over 24 timer', 'Koordinationsplan',
    'Udskrivningsaftale', 'Særlige dørlåse', 'Personlige alarm- og pejlesystemer', 'Andet' )"""

    view = "[FOR_tvang_alt_hele_kohorten_inkl_2021_feb2022]"

    sql = f"SELECT dw_ek_borger, datotid_start_sei, datotid_slut_sei, varighed_timer_sei, typetekst_sei FROM [fct].{view} WHERE datotid_start_sei IS NOT NULL AND typetekst_sei NOT IN {coercion_discard}"

    if coercion_type and reason_for_coercion is None:

        sql += f" AND typetekst_sei = '{coercion_type}'"

    if coercion_type is None and reason_for_coercion:

        sql += f" AND begrundtekst_sei = '{reason_for_coercion}'"

    if coercion_type and reason_for_coercion:

        sql += f" AND typetekst_sei = '{coercion_type}' AND begrundtekst_sei = '{reason_for_coercion}'"

    df = sql_load(sql, database="USR_PS_FORSK", chunksize=None, n_rows=n_rows)

    # add end time as start time for acute sedation
    df.loc[df.typetekst_sei == "Beroligende medicin", "datotid_slut_sei"] = df[
        "datotid_start_sei"
    ]

    # drop nas for coercion end times
    df = df.dropna(subset="datotid_slut_sei")

    # Drop duplicate rows
    df = df.drop_duplicates(keep="first")

    df.rename(
        columns={"datotid_slut_sei": "timestamp", "varighed_timer_sei": "value"},
        inplace=True,
    )

    # Change NaNs to 0
    df["value"].fillna(0, inplace=True)

    return df[["dw_ek_borger", "timestamp", "value"]].reset_index(drop=True)


def _concatenate_coercion(
    coercion_types_list: list[dict[str, str]],
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Aggregate multiple types of coercion with multiple reasons into one
    column.

    Args:
        coercion_types_list (list): list of dictionaries containing a 'coercion_type' key and a 'reason_for_coercion' key. If keys not in dicts, they are set to None # noqa: DAR102
        n (int, optional): Number of rows to return. Defaults to None.

    Returns:
        pd.DataFrame
    """

    for d in coercion_types_list:  # Make sure proper keys are given
        if "coercion_type" not in d and "reason_for_coercion" not in d:
            raise KeyError(
                f'{d} does not contain either "coercion_type"  or "reason_for_coercion". At least one is required.',
            )
        if "coercion_type" not in d:
            d["coercion_type"] = None  # type: ignore
        if "reason_for_coercion" not in d:
            d["reason_for_coercion"] = None  # type: ignore

    dfs = [
        coercion_duration(
            coercion_type=d["coercion_type"],
            reason_for_coercion=d["reason_for_coercion"],
            n_rows=n_rows,
        )
        for d in coercion_types_list
    ]

    return pd.concat(dfs, axis=0).reset_index(drop=True)


# REASON (begrundtekst_sei) #


@data_loaders.register("farlighed")
def farlighed(n_rows: Optional[int] = None) -> pd.DataFrame:
    coercion_types_list = [
        {
            "reason_for_coercion": "Farlighed",
        },
        {
            "reason_for_coercion": "På grund af farlighed",
        },
    ]

    return _concatenate_coercion(
        coercion_types_list=coercion_types_list,
        n_rows=n_rows,
    )


# Røde papir ved tvangsindlæggelse/tvangstilbageholdelse
@data_loaders.register("paa_grund_af_farlighed")
def paa_grund_af_farlighed(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        reason_for_coercion="På grund af farlighed",
        n_rows=n_rows,
    )


# Gule papir ved tvangsindlæggelse/tvangstilbageholdelse
@data_loaders.register("af_helbredsmaessige_grunde")
def af_helbredsmaessige_grunde(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        reason_for_coercion=" Af helbredsmæssige grunde",
        n_rows=n_rows,
    )


@data_loaders.register("urolig_tilstand")
def urolig_tilstand(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        reason_for_coercion="Urolig tilstand",
        n_rows=n_rows,
    )


@data_loaders.register("anden_begrundelse")
def anden_begrundelse(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        reason_for_coercion="Anden begrundelse",
        n_rows=n_rows,
    )


@data_loaders.register("naerliggende_eller_vaesentlig_fare_for_patienten_eller_andre")
def naerliggende_fare(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        reason_for_coercion="Nærliggende_eller_væsentlig_fare_for_patienten_eller_andre",
        n_rows=n_rows,
    )


# GENERAL TYPE (tabeltekst) ###
# frihedsberøvelser
@data_loaders.register("skema_1")
def skema_1(n_rows: Optional[int] = None) -> pd.DataFrame:
    coercion_types_list = [
        {
            "coercion_type": "Tvangsindlæggelse",
        },
        {
            "coercion_type": "Tvangstilbageholdelse",
        },
    ]

    return _concatenate_coercion(
        coercion_types_list=coercion_types_list,
        n_rows=n_rows,
    )


# tvangsbehandlinger
@data_loaders.register("skema_2")
def skema_2(n_rows: Optional[int] = None) -> pd.DataFrame:
    coercion_types_list = [
        {
            "coercion_type": "Af legemlig lidelse",
        },
        {
            "coercion_type": "Medicinering",
        },
        {
            "coercion_type": "Ernæring",
        },
        {
            "coercion_type": "ECT",
        },
    ]

    return _concatenate_coercion(
        coercion_types_list=coercion_types_list,
        n_rows=n_rows,
    )


@data_loaders.register("skema_2_without_nutrition")
def skema_2_without_nutrition(n_rows: Optional[int] = None) -> pd.DataFrame:
    coercion_types_list = [
        {
            "coercion_type": "Af legemlig lidelse",
        },
        {
            "coercion_type": "Medicinering",
        },
        {
            "coercion_type": "ECT",
        },
    ]

    return _concatenate_coercion(
        coercion_types_list=coercion_types_list,
        n_rows=n_rows,
    )


# magtanvendelse
@data_loaders.register("skema_3")
def skema_3(n_rows: Optional[int] = None) -> pd.DataFrame:
    coercion_types_list = [
        {
            "coercion_type": "Bælte",
        },
        {
            "coercion_type": "Remme",
        },
        {
            "coercion_type": "Fastholden",
        },
        {
            "coercion_type": "Beroligende medicin",
        },
        {
            "coercion_type": "Handsker",
        },
    ]

    # "døraflåsning" and "personlig skærmning" are not included

    return _concatenate_coercion(
        coercion_types_list=coercion_types_list,
        n_rows=n_rows,
    )


# SPECIFIC TYPE (typetekst_sei) ###
# exists in the data, but not included here: [døraflåsning, personlig afskærmning, stofbælte, særlige dørlåse, tvungen opfølgning, personlige alarm, udskrivningsaftale, koordinationsplan]


@data_loaders.register("baelte")
def baelte(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Bælte",
        n_rows=n_rows,
    )


@data_loaders.register("remme")
def remme(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Remme",
        n_rows=n_rows,
    )


@data_loaders.register("fastholden")
def fastholden(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Fastholden",
        n_rows=n_rows,
    )


@data_loaders.register("beroligende_medicin")
def beroligende_medicin(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Beroligende medicin",
        n_rows=n_rows,
    )


@data_loaders.register("handsker")
def handsker(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Handsker",
        n_rows=n_rows,
    )


@data_loaders.register("tvangsindlaeggelse")
def tvangsindlaeggelse(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Tvangsindlæggelse",
        n_rows=n_rows,
    )


@data_loaders.register("tvangstilbageholdelse")
def tvangstilbageholdelse(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Tvangstilbageholdelse",
        n_rows=n_rows,
    )


@data_loaders.register("medicinering")
def medicinering(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Medicinering",
        n_rows=n_rows,
    )


@data_loaders.register("ect")
def ect(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="ECT",
        n_rows=n_rows,
    )


@data_loaders.register("ernaering")
def ernaering(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Ernæring",
        n_rows=n_rows,
    )


@data_loaders.register("af_legemlig_lidelse")
def af_legemlig_lidelse(n_rows: Optional[int] = None) -> pd.DataFrame:
    return coercion_duration(
        coercion_type="Af legemlig lidelse",
        n_rows=n_rows,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['psycop_feature_generation',
 'psycop_feature_generation.application_modules',
 'psycop_feature_generation.data_checks',
 'psycop_feature_generation.data_checks.flattened',
 'psycop_feature_generation.data_checks.raw',
 'psycop_feature_generation.featurizers',
 'psycop_feature_generation.loaders',
 'psycop_feature_generation.loaders.example',
 'psycop_feature_generation.loaders.flattened',
 'psycop_feature_generation.loaders.raw',
 'psycop_feature_generation.loaders.synth.raw']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.41,<1.5.42',
 'catalogue>=2.0.0,<2.1.0',
 'dask>=2022.9.0,<2022.13.0',
 'deepchecks>=0.8.0,<0.11.0',
 'dill>=0.3.0,<0.3.6',
 'frozendict>=2.3.4,<3.0.0',
 'numpy>=1.23.3,<1.23.6',
 'pandas>=1.4.0,<1.6.0',
 'protobuf<=3.20.3',
 'psutil>=5.9.1,<6.0.0',
 'psycopmlutils>=0.2.4,<0.3.0',
 'pyarrow>=9.0.0,<9.1.0',
 'pydantic>=1.9.0,<1.10.0',
 'pyodbc>=4.0.34,<4.0.36',
 'scikit-learn>=1.1.2,<1.1.3',
 'scipy>=1.8.0,<1.9.4',
 'srsly>=2.4.4,<2.4.6',
 'timeseriesflattener>=0.21.0',
 'transformers>=4.22.2,<5.0.0',
 'wandb>=0.12.0,<0.13.5',
 'wasabi>=0.9.1,<0.10.2']

setup_kwargs = {
    'name': 'psycop-feature-generation',
    'version': '0.17.1',
    'description': '',
    'long_description': 'None',
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



"""Loaders for diagnosis codes.

Is growing quite a bit, loaders may have to be split out into separate
files (e.g. psychiatric, cardiovascular, metabolic etc.) over time.
"""

# pylint: disable=missing-function-docstring

import logging
from typing import Optional, Union

import pandas as pd

from psycop_feature_generation.loaders.raw.utils import load_from_codes
from psycop_feature_generation.utils import data_loaders

log = logging.getLogger(__name__)


def from_contacts(
    icd_code: Union[list[str], str],
    output_col_name: Optional[str] = "value",
    n_rows: Optional[int] = None,
    wildcard_icd_code: Optional[bool] = False,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    """Load diagnoses from all hospital contacts. If icd_code is a list, will
    aggregate as one column (e.g. ["E780", "E785"] into a ypercholesterolemia
    column).

    Args:
        icd_code (str): Substring to match diagnoses for. Matches any diagnoses, whether a-diagnosis, b-diagnosis etc. # noqa: DAR102
        output_col_name (str, optional): Name of new column string. Defaults to "value".
        n_rows: Number of rows to return. Defaults to None.
        wildcard_icd_code (bool, optional): Whether to match on icd_code*. Defaults to False.
        shak_location_col (str, optional): Name of column containing shak code. Defaults to None. For diagnosis loaders, this column is "shakkode_ansvarlig". Combine with shak_code and shak_sql_operator.
        shak_code (int, optional): Shak code indicating where to keep/not keep visits from (e.g. 6600). Defaults to None.
        shak_sql_operator (str, optional): Operator indicating how to filter shak_code, e.g. "!= 6600" or "= 6600". Defaults to None.

    Returns:
        pd.DataFrame
    """

    log.warning(
        "The DNPR3 data model replaced the DNPR2 model on 3 February 2019. "
        "Due to changes in DNPR3, granurality of diagnoses differ across the two models. "
        "If your prediction timestamps, lookbehind or lookahead span across this date, "
        "you should not use count as a resolve_multiple_fn. "
        "See the wiki (LPR2 compared to LPR3) for more information.",
    )

    df = load_from_codes(
        codes_to_match=icd_code,
        code_col_name="diagnosegruppestreng",
        source_timestamp_col_name="datotid_slut",
        view="FOR_kohorte_indhold_pt_journal_psyk_somatik_inkl_2021_feb2022",
        output_col_name=output_col_name,
        match_with_wildcard=wildcard_icd_code,
        n_rows=n_rows,
        load_diagnoses=True,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )

    df = df.drop_duplicates(
        subset=["dw_ek_borger", "timestamp", output_col_name],
        keep="first",
    )

    return df.reset_index(drop=True)


@data_loaders.register("essential_hypertension")
def essential_hypertension(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="I109",
        wildcard_icd_code=False,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("hyperlipidemia")
def hyperlipidemia(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=[
            "E780",
            "E785",
        ],  # Only these two, as the others are exceedingly rare
        wildcard_icd_code=False,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("liverdisease_unspecified")
def liverdisease_unspecified(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="K769",
        wildcard_icd_code=False,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("polycystic_ovarian_syndrome")
def polycystic_ovarian_syndrome(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="E282",
        wildcard_icd_code=False,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("sleep_apnea")
def sleep_apnea(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["G473", "G4732"],
        wildcard_icd_code=False,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("sleep_problems_unspecified")
def sleep_problems_unspecified(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="G479",
        wildcard_icd_code=False,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("copd")
def copd(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["j44"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# Psychiatric diagnoses
# data loaders for all diagnoses in the f0-chapter (organic mental disorders)
@data_loaders.register("f0_disorders")
def f0_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f0",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("dementia")
def dementia(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f00", "f01", "f02", "f03", "f04"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("delirium")
def delirium(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f05",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("miscellaneous_organic_mental_disorders")
def misc_organic_mental_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f06", "f07", "f09"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# data loaders for all diagnoses in the f1-chapter (mental and behavioural disorders due to psychoactive substance use)
@data_loaders.register("f1_disorders")
def f1_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f1",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("alcohol_dependency")
def alcohol_dependency(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f10",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("opioid_dependency")
def opioids_and_sedatives(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f11",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("cannabinoid_dependency")
def cannabinoid_dependency(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f12",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("sedative_dependency")
def sedative_dependency(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f13",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("stimulant_dependencies")
def stimulant_deo(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f14", "f15"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("hallucinogen_dependency")
def hallucinogen_dependency(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f16",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("tobacco_dependency")
def tobacco_dependency(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f17",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("miscellaneous_drug_dependencies")
def misc_drugs(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f18", "f19"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# data loaders for all diagnoses in the f2-chapter (schizophrenia, schizotypal and delusional disorders)


@data_loaders.register("f2_disorders")
def f2_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f2",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("schizophrenia")
def schizophrenia(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f20",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("schizoaffective")
def schizoaffective(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f25",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("miscellaneous_psychotic_disorders")
def misc_psychosis(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f21", "f22", "f23", "f24", "f28", "f29"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# data loaders for all diagnoses in the f3-chapter (mood (affective) disorders).


@data_loaders.register("f3_disorders")
def f3_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f3",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("manic_and_bipolar")
def manic_and_bipolar(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f30", "f31"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("depressive_disorders")
def depressive_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f32", "f33", "f34", "f38"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("miscellaneous_affective_disorders")
def misc_affective_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f38", "f39"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# data loaders for all diagnoses in the f4-chapter (neurotic, stress-related and somatoform disorders).


@data_loaders.register("f4_disorders")
def f4_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f4",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("phobic_anxiety_ocd")
def phobic_and_anxiety(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f40", "f41", "f42"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("reaction_to_severe_stress_and_adjustment_disorders")
def stress_and_adjustment(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f43",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("dissociative_somatoform_miscellaneous")
def dissociative_somatoform_and_misc(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f44", "f45", "f48"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# data loaders for all diagnoses in the f5-chapter (behavioural syndromes associated with physiological disturbances and physical factors).


@data_loaders.register("f5_disorders")
def f5_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f5",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("eating_disorders")
def eating_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f50",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("sleeping_and_sexual_disorders")
def sleeping_and_sexual_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f51", "f52"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("miscellaneous_f5_disorders")
def misc_f5(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f53", "f54", "f55", "f59"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# data loaders for all diagnoses in the f6-chapter (disorders of adult personality and behaviour).
@data_loaders.register("f6_disorders")
def f6_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f6",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("cluster_a")
def cluster_a(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f600", "f601"],
        wildcard_icd_code=False,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("cluster_b")
def cluster_b(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f602", "f603", "f604"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("cluster_c")
def cluster_c(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f605", "f606", "f607"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("miscellaneous_personality_disorders")
def misc_personality_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f608", "f609", "f61", "f62", "f63", "f68", "f69"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("sexual_disorders")
def misc_personality(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f65", "f66"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )

    # f64 sexual identity disorders is excluded


# data loaders for all diagnoses in the f7-chapter (mental retardation).
@data_loaders.register("f7_disorders")
def f7_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f7",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("mild_mental_retardation")
def mild_mental_retardation(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f70",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("moderate_mental_retardation")
def moderate_mental_retardation(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f71",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("severe_mental_retardation")
def severe_mental_retardation(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f72", "f73"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("miscellaneous_mental_retardation_disorders")
def misc_mental_retardation(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f78", "f79"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# data loaders for all diagnoses in the f8-chapter (disorders of psychological development).
@data_loaders.register("f8_disorders")
def f8_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f8",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("pervasive_developmental_disorders")
def pervasive_developmental_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f84",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("miscellaneous_f8_disorders")
def misc_f8(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f80", "f81", "f82", "f83", "f88", "f89"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


# data loaders for all diagnoses in the f9-chapter (child and adolescent disorders).
@data_loaders.register("f9_disorders")
def f9_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f9",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("hyperkinetic_disorders")
def hyperkinetic_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code="f90",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("behavioural_disorders")
def behavioural_disorders(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f91", "f92", "f93", "f94"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("tics_and_miscellaneous_f9")
def tics_and_misc(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    return from_contacts(
        icd_code=["f95", "f98"],
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("gerd")
def gerd(
    n_rows: Optional[int] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    """Gastroesophageal reflux disease (GERD) diagnoses."""
    return from_contacts(
        icd_code="k21",
        wildcard_icd_code=True,
        n_rows=n_rows,
        shak_location_col=shak_location_col,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )



"""Generates a df with feature descriptions for the predictors in the source
df."""
from collections.abc import Sequence
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from timeseriesflattener.feature_spec_objects import (
    PredictorSpec,
    StaticSpec,
    TemporalSpec,
    _AnySpec,
)
from wasabi import Printer

from psycop_feature_generation.data_checks.utils import save_df_to_pretty_html_table
from psycop_feature_generation.loaders.flattened.local_feature_loaders import (
    load_split_predictors,
)

UNICODE_HIST = {
    0: " ",
    1 / 8: "▁",
    1 / 4: "▂",
    3 / 8: "▃",
    1 / 2: "▄",
    5 / 8: "▅",
    3 / 4: "▆",
    7 / 8: "▇",
    1: "█",
}

HIST_BINS = 8


def get_value_proportion(series, value):
    """Get proportion of series that is equal to the value argument."""
    if np.isnan(value):
        return round(series.isna().mean(), 2)
    else:
        return round(series.eq(value).mean(), 2)


def _find_nearest(array, value):
    """Find the nearest numerical match to value in an array.

    Args:
        array (np.ndarray): An array of numbers to match with.
        value (float): Single value to find an entry in array that is close.

    Returns:
        np.array: The entry in array that is closest to value.
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def create_unicode_hist(series: pd.Series) -> pd.Series:
    """Return a histogram rendered in block unicode. Given a pandas series of
    numerical values, returns a series with one entry, the original series
    name, and a histogram made up of unicode characters.

    Args:
        series (pd.Series): Numeric column of data frame for analysis

    Returns:
        pd.Series: Index of series name and entry with unicode histogram as
        a string, eg '▃▅█'

    All credit goes to the python package skimpy.
    """
    # Remove any NaNs
    series = series.dropna()

    if series.dtype == "bool":
        series = series.astype("int")

    hist, _ = np.histogram(series, density=True, bins=HIST_BINS)
    hist = hist / hist.max()

    # Now do value counts
    key_vector = np.array(list(UNICODE_HIST.keys()), dtype="float")

    ucode_to_print = "".join(
        [UNICODE_HIST[_find_nearest(key_vector, val)] for val in hist],
    )

    return ucode_to_print


def generate_temporal_feature_description(
    series: pd.Series,
    predictor_spec: TemporalSpec,
):
    """Generate a row with feature description for a temporal predictor."""
    d = {
        "Predictor df": predictor_spec.feature_name,
        "Lookbehind days": predictor_spec.interval_days,
        "Resolve multiple": predictor_spec.resolve_multiple_fn,
        "N unique": series.nunique(),
        "Fallback strategy": predictor_spec.fallback,
        "Proportion missing": series.isna().mean(),
        "Mean": round(series.mean(), 2),
        "Histogram": create_unicode_hist(series),
        "Proportion using fallback": get_value_proportion(
            series,
            predictor_spec.fallback,
        ),
    }

    for percentile in (0.01, 0.25, 0.5, 0.75, 0.99):
        # Get the value representing the percentile
        d[f"{percentile * 100}-percentile"] = round(series.quantile(percentile), 1)

    return d


def generate_static_feature_description(series: pd.Series, predictor_spec: StaticSpec):
    """Generate a row with feature description for a static predictor."""
    return {
        "Predictor df": predictor_spec.feature_name,
        "Lookbehind days": "N/A",
        "Resolve multiple": "N/A",
        "N unique": series.nunique(),
        "Fallback strategy": "N/A",
        "Proportion missing": series.isna().mean(),
        "Mean": round(series.mean(), 2),
        "Histogram": create_unicode_hist(series),
        "Proportion using fallback": "N/A",
    }


def generate_feature_description_row(
    series: pd.Series,
    predictor_spec: _AnySpec,
) -> dict:
    """Generate a row with feature description.

    Args:
        series (pd.Series): Series with data to describe.
        predictor_spec (PredictorSpec): Predictor specification.

    Returns:
        dict: dictionary with feature description.
    """

    if isinstance(predictor_spec, StaticSpec):
        d = generate_static_feature_description(series, predictor_spec)
    elif isinstance(predictor_spec, TemporalSpec):
        d = generate_temporal_feature_description(series, predictor_spec)

    return d


def generate_feature_description_df(
    df: pd.DataFrame,
    predictor_specs: list[PredictorSpec],
) -> pd.DataFrame:
    """Generate a data frame with feature descriptions.

    Args:
        df (pd.DataFrame): Data frame with data to describe.
        predictor_specs (PredictorSpec): Predictor specifications.

    Returns:
        pd.DataFrame: Data frame with feature descriptions.
    """

    rows = []

    for spec in predictor_specs:
        column_name = spec.get_col_str()

        rows.append(
            generate_feature_description_row(
                series=df[column_name],
                predictor_spec=spec,
            ),
        )

    # Convert to dataframe
    feature_description_df = pd.DataFrame(rows)

    # Sort feature_description_df by Predictor df to make outputs easier to read
    feature_description_df = feature_description_df.sort_values(by="Predictor df")

    return feature_description_df


def save_feature_description_from_dir(
    feature_set_dir: Path,
    feature_specs: list[Union[TemporalSpec, StaticSpec]],
    file_suffix: str,
    splits: Sequence[str] = ("train",),
    out_dir: Path = None,
):
    """Write a csv with feature descriptions in the directory.

    Args:
        feature_set_dir (Path): Path to directory with data frames.
        feature_specs (list[PredictorSpec]): List of feature specifications.
        file_suffix (str): Suffix of the data frames to load. Must be either ".csv" or ".parquet".
        splits (tuple[str]): tuple of splits to include in the description. Defaults to ("train").
        out_dir (Path): Path to directory where to save the feature description. Defaults to None.
    """
    msg = Printer(timestamp=True)

    if out_dir is None:
        save_dir = feature_set_dir / "feature_descriptions"

    else:
        save_dir = out_dir / "feature_descriptions"

    if not save_dir.exists():
        save_dir.mkdir()

    for split in splits:
        msg.info(f"{split}: Creating feature description")

        predictors = load_split_predictors(
            feature_set_dir=feature_set_dir,
            split=split,
            include_id=False,
            file_suffix=file_suffix,
        )

        msg.info(f"{split}: Generating feature description dataframe")

        feature_description_df = generate_feature_description_df(
            df=predictors,
            predictor_specs=feature_specs,
        )

        msg.info(f"{split}: Writing feature description to disk")

        feature_description_df.to_csv(
            save_dir / f"{split}_feature_description.csv",
            index=False,
        )
        # Writing html table as well
        save_df_to_pretty_html_table(
            path=save_dir / f"{split}_feature_description.html",
            title="Feature description",
            df=feature_description_df,
        )



"""Wandb utils.""" ""
import traceback

import wandb


def wandb_alert_on_exception(func):
    """Alerts wandb on exception."""

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            wandb.alert(title="Run crashed", text=traceback.format_exc())
            raise e

    return wrapper



"""Utility functions for the application modules."""
import functools
import logging
from typing import Callable

import pandas as pd


def print_df_dimensions_diff(
    func: Callable,
    print_when_starting: bool = False,
    print_when_no_diff=False,
):
    """Print the difference in rows between the input and output dataframes."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        log = logging.getLogger(__name__)

        if print_when_starting:
            log.info(f"{func.__name__}: Starting")

        arg_dfs = [arg for arg in args if isinstance(arg, pd.DataFrame)]
        kwargs_dfs = [arg for arg in kwargs.values() if isinstance(arg, pd.DataFrame)]
        potential_dfs = arg_dfs + kwargs_dfs

        if len(potential_dfs) > 1:
            raise ValueError("More than one DataFrame found in args or kwargs")

        df = potential_dfs[0]

        for dim in ("rows", "columns"):
            dim_int = 0 if dim == "rows" else 1

            n_in_dim_before_func = df.shape[dim_int]

            if print_when_no_diff:
                log.info(
                    f"{func.__name__}: {n_in_dim_before_func} {dim} before function",
                )

            result = func(*args, **kwargs)

            diff = n_in_dim_before_func - result.shape[dim_int]

            if diff != 0:
                percent_diff = round(
                    (n_in_dim_before_func - result.shape[dim_int])
                    / n_in_dim_before_func
                    * 100,
                    2,
                )

                log.info(f"{func.__name__}: Dropped {diff} ({percent_diff}%) {dim}")
            else:
                if print_when_no_diff:
                    log.info(f"{func.__name__}: No {dim} dropped")

        return result

    return wrapper



"""Loader for moves between regions."""
from typing import Optional

import pandas as pd

from psycop_feature_generation.loaders.raw.sql_load import sql_load


def load_moves(n_rows: Optional[int] = None) -> pd.DataFrame:
    """Get a dataframe with timestamps for each move into or out from the
    Central Denmark Region."""
    view = "[bopael_i_rm]"

    sql = f"SELECT * FROM [fct].{view}"

    df = sql_load(sql, database="USR_PS_FORSK", chunksize=None, n_rows=n_rows)

    return df


def load_move_into_rm_for_exclusion(n_rows: Optional[int] = None) -> pd.DataFrame:
    """Get a dataframe with timestamps for each move into or out from the
    Central Denmark Region."""
    df = load_moves(n_rows=n_rows)

    df = df.drop(columns=["fraflytnings_datotid"])
    df = df.rename(columns={"tilflytnings_datotid": "timestamp"})
    df = df.loc[df["timestamp"] >= pd.to_datetime("2012-01-01")]

    return df


if __name__ == "__main__":
    df = load_move_into_rm_for_exclusion()



"""A set of misc.

utilities. If this file grows, consider splitting it up.
"""

import os
from collections.abc import Hashable
from pathlib import Path
from typing import Any, Optional

import catalogue
import pandas as pd

data_loaders = catalogue.create("timeseriesflattener", "data_loaders")

SHARED_RESOURCES_PATH = Path(r"E:\shared_resources")
RELATIVE_PROJECT_ROOT = Path(__file__).resolve().parents[2]


def format_dict_for_printing(d: dict) -> str:
    """Format a dictionary for printing. Removes extra apostrophes, formats
    colon to dashes, separates items with underscores and removes curly
    brackets.

    Args:
        d (dict): dictionary to format.

    Returns:
        str: Formatted dictionary.

    Example:
        >>> d = {"a": 1, "b": 2}
        >>> print(format_dict_for_printing(d))
        >>> "a-1_b-2"
    """
    return (
        str(d)
        .replace("'", "")
        .replace(": ", "-")
        .replace("{", "")
        .replace("}", "")
        .replace(", ", "_")
    )


def load_dataset_from_file(
    file_path: Path,
    nrows: Optional[int] = None,
) -> pd.DataFrame:
    """Load dataset from file. Handles csv and parquet files based on suffix.

    Args:
        file_path (str): File name.
        nrows (int): Number of rows to load.

    Returns:
        pd.DataFrame: Dataset
    """

    file_suffix = file_path.suffix

    if file_suffix == ".csv":
        return pd.read_csv(file_path, nrows=nrows)

    elif file_suffix == ".parquet":
        if nrows:
            raise ValueError("nrows not supported for parquet files")

        return pd.read_parquet(file_path)
    else:
        raise ValueError(f"Invalid file suffix {file_suffix}")


def load_most_recent_file_matching_pattern_as_df(
    dir_path: Path,
    file_pattern: str,
    file_suffix: str,
) -> pd.DataFrame:
    """Load most recent df matching pattern.

    Args:
        dir_path (Path): Directory to search
        file_pattern (str): Pattern to match
        file_suffix (str): File suffix. Must be either ".csv" or ".parquet".

    Returns:
        pd.DataFrame: DataFrame matching pattern

    Raises:
        FileNotFoundError: If no file matching pattern is found
    """
    files = list(dir_path.glob(f"*{file_pattern}*.{file_suffix}"))

    if len(files) == 0:
        raise FileNotFoundError(f"No files matching pattern {file_pattern} found")

    most_recent_file = max(files, key=os.path.getctime)

    return load_dataset_from_file(file_path=most_recent_file)


def df_contains_duplicates(df: pd.DataFrame, col_subset: list[str]):
    """Check if a dataframe contains duplicates.

    Args:
        df (pd.DataFrame): Dataframe to check.
        col_subset (list[str]): Columns to check for duplicates.

    Returns:
        bool: True if duplicates are found.
    """
    return df.duplicated(subset=col_subset).any()


def write_df_to_file(
    df: pd.DataFrame,
    file_path: Path,
):
    """Write dataset to file. Handles csv and parquet files based on suffix.

    Args:
        df: Dataset
        file_path (str): File name.
    """

    file_suffix = file_path.suffix

    if file_suffix == ".csv":
        df.to_csv(file_path, index=False)
    elif file_suffix == ".parquet":
        df.to_parquet(file_path, index=False)
    else:
        raise ValueError(f"Invalid file suffix {file_suffix}")


def assert_no_duplicate_dicts_in_list(predictor_spec_list: list[dict[str, Any]]):
    """Find potential duplicates in list of dicts.

    Args:
        predictor_spec_list (list[dict[str, dict[str, Any]]]): List of predictor combinations.
    """
    # Find duplicates in list of dicts
    seen = set()
    duplicates = set()

    for d in predictor_spec_list:
        # Remove any keys with unhashable values
        # Otherwise, we get an error when using "in".
        d = {k: v for k, v in d.items() if isinstance(v, Hashable)}

        d_as_tuple = tuple(d.items())
        if d_as_tuple in seen:  # pylint: disable=R6103
            duplicates.add(d_as_tuple)
        else:
            seen.add(d_as_tuple)

    if len(duplicates) > 0:
        raise ValueError(f"Found duplicates in list of dicts: {duplicates}")
